--=============================================================================--
--===========================================================================--
library ieee;
   use ieee.std_logic_1164.all;
   use IEEE.STD_LOGIC_ARITH.ALL;
   use IEEE.STD_LOGIC_UNSIGNED.ALL;
   use ieee.numeric_std.all;
	
use work.pkg_vectrex.all;

entity vectrex is
  port(
		cpu_areset   : in    std_logic;
		cpu_ahold    : in    std_logic;
		cpu_anmi     : in    std_logic;
		cpu_clk		: in    std_logic; --12 MHz
		analog_sound_out    : out std_logic_vector(7 downto 0);
		analog_x_out    : out std_logic_vector(9 downto 0);
		analog_y_out    : out std_logic_vector(9 downto 0);
		analog_z_out    : out std_logic_vector(7 downto 0);
		p1_x				 : in std_logic_vector(7 downto 0);
		p1_y				 : in std_logic_vector(7 downto 0);
		p1_buttons		 : in std_logic_vector(3 downto 0);
		p2_x				 : in std_logic_vector(7 downto 0);
		p2_y				 : in std_logic_vector(7 downto 0);
		p2_buttons		 : in std_logic_vector(3 downto 0)
	);
end vectrex;

architecture Behaviour of vectrex is
  type hold_state_type is ( hold_release_state, hold_request_state );

  signal sys_clk       	: std_logic; 
  signal vga_clk       	: std_logic;
  signal pll_locked    	: std_logic;
  signal resetsr 		: std_logic_vector(1 downto 0);
  signal ares			: std_logic;
  signal cnt1m5			: integer range 0 to 7; 
  signal clk1m5			: std_logic;
  signal cnt6m			: integer range 0 to 1; 
  signal clk6m			: std_logic;
  signal cnt750k		: integer range 0 to 7; 
  signal clk750k		: std_logic;
  
  -- CPU Interface signals
  signal cpu_rst       : std_logic;
  signal cpu_rw        : std_logic;
  signal cpu_vma       : std_logic;
  signal cpu_halt      : std_logic;
  signal cpu_hold      : std_logic;
  signal cpu_nmi      : std_logic;
  signal cpu_firq      : std_logic;
  signal cpu_irq       : std_logic;
  signal cpu_addr      : std_logic_vector(15 downto 0);
  signal cpu_data_in   : std_logic_vector(7 downto 0);
  signal cpu_data_out  : std_logic_vector(7 downto 0);
  signal cpu_pc		   : std_logic_vector(15 downto 0);

  -- BOOT ROM
  signal rom_cs        : Std_logic;
  signal rom_data_out  : Std_Logic_Vector(7 downto 0);

  -- IO / VIA
  signal io_cs       : std_logic;
  signal io_data_out : std_logic_vector(7 downto 0);
  signal io_data_oe  : std_logic;
  signal via_wen	 : std_logic;
  signal via_ren     : std_logic;
  signal via_pa_out  : std_logic_vector(7 downto 0);
  signal via_pa_oe	 : std_logic_vector(7 downto 0);
  signal via_pa_in   : std_logic_vector(7 downto 0);
  signal via_pb_out  : std_logic_vector(7 downto 0);
  signal via_pb_oe	 : std_logic_vector(7 downto 0);
  signal via_pb_in	 : std_logic_vector(7 downto 0);
  signal via_irq_out : std_logic; -- ?????????????
  signal via_ca1_in  : std_logic; -- to IO7 AY8912
  signal via_ca2_out : std_logic; -- alg n_zero
  signal via_ca2_oe  : std_logic; -- alg n_zero
  signal via_ca2_in  : std_logic; -- alg n_zero
  signal via_cb1_out : std_logic; -- nc
  signal via_cb1_oe  : std_logic; -- nc
  signal via_cb1_in  : std_logic; -- nc
  signal via_cb2_out : std_logic; -- alg n_blank
  signal via_cb2_oe  : std_logic; -- alg n_blank
  signal via_cb2_in  : std_logic; -- alg n_blank
  
  -- Sound / AY
  signal ay_bc1		 : std_logic; -- to AY bc1
  signal ay_bdir     : std_logic; -- to AY bdir
  signal ay_pa_out	 : std_logic_vector(7 downto 0);
  signal ay_pa_in 	 : std_logic_vector(7 downto 0);
  signal ay_data_out : std_logic_vector(7 downto 0);
  signal ay_data_in  : std_logic_vector(7 downto 0);
  signal ay_data_oe  : std_logic;
  
  -- CART
  signal cart_cs       : std_logic;
  signal cart_data_out : std_logic_vector(7 downto 0);

  -- RAM
  signal sram_cs       : std_logic;
  signal sram_data_out : std_logic_vector(7 downto 0);
  signal sram_ce       : std_logic;
  signal sram_ub       : std_logic;
  signal sram_lb       : std_logic;
  signal sram_we       : std_logic;
  signal sram_oe       : std_logic;
  
  --Analog
  signal analog_vec    : std_logic_vector(7 downto 0);
  signal analog_x      : std_logic_vector(7 downto 0);
  signal analog_y      : std_logic_vector(7 downto 0);
  signal analog_y_temp : std_logic_vector(7 downto 0);
  signal analog_x_int  : std_logic_vector(9 downto 0);
  signal analog_y_int  : std_logic_vector(9 downto 0);
  signal analog_z      : std_logic_vector(7 downto 0);
  signal analog_z_sh   : std_logic_vector(7 downto 0);
  signal analog_sound  : std_logic_vector(7 downto 0);
  signal analog_offs   : std_logic_vector(7 downto 0);
  signal comparator_in : std_logic_vector(7 downto 0);
  signal compare_out   : std_logic;
  signal cs_x          : std_logic;
  signal cs_y          : std_logic;
  signal cs_x_int      : std_logic;
  signal cs_y_int      : std_logic;
  signal cs_z          : std_logic;


begin
  -----------------------------------------------------------------------------
  -- Instantiation of internal components
  -----------------------------------------------------------------------------

my_cpu : cpu09  port map (    
	 clk	   => not clk6m,
	 rst       => cpu_rst,
     vma       => cpu_vma,
     addr      => cpu_addr(15 downto 0),
     rw	       => cpu_rw,
     data_in   => cpu_data_in,
	 data_out  => cpu_data_out,
	 halt      => '0',--cpu_halt,
	 hold      => '0',--cpu_hold,
	 irq       => '0', --not via_irq_out,
	 nmi       => '0', --cpu_nmi,
	 firq      => '0',--ay_pa_out(3),--cpu_firq,
	 dpc	   => cpu_pc
    );

my_rom : mon_rom port map (
    clk       => clk6m,
    rst       => cpu_rst,
	cs        => rom_cs,
	rw        => '1',
    addr      => cpu_addr(12 downto 0),
    data_in   => cpu_data_out,
    data_out  => rom_data_out
    );
    
my_cart : cart_rom port map (
    clk       => clk6m,
    rst       => cpu_rst,
	cs        => cart_cs,
	rw        => '1',
    addr      => cpu_addr(12 downto 0),
    data_in   => cpu_data_out,
    data_out  => cart_data_out
    );
    
my_ram : ram_2k port map (
	clk       => clk6m,
    rst       => cpu_rst,
	cs        => sram_cs,
	rw        => cpu_rw,
    addr      => cpu_addr(9 downto 0),
    data_in   => cpu_data_out,
    data_out  => sram_data_out
	);


my_via : via6522 port map (
    
		clock       => clk6m,
        clock_en    => clk1m5,
        reset       => cpu_rst,
                                
        addr        => cpu_addr(3 downto 0),
        wen         => via_wen,
        ren         => via_ren,
        data_in     => cpu_data_out,
        data_out    => io_data_out,
                                
        -- pio --   
        port_a_o    => via_pa_out,
        port_a_t    => via_pa_oe,
        port_a_i    => via_pa_in,
                                
        port_b_o    => via_pb_out,
        port_b_t    => via_pb_oe,
        port_b_i    => via_pb_in,
    
        -- handshake pins
        ca1_i       => ay_pa_out(7),--via_ca1_in,
                            
        ca2_o       => via_ca2_out,
        ca2_i       => via_ca2_in,
        ca2_t       => via_ca2_oe,
                            
        cb1_o       => via_cb1_out,
        cb1_i       => via_cb1_in,
        cb1_t       => via_cb1_oe,
                            
        cb2_o       => via_cb2_out,
        cb2_i       => via_cb2_in,
        cb2_t       => via_cb2_oe,
                            
        irq         => via_irq_out
    );

    
my_ay : YM2149
	port map (
	-- data bus
    I_DA                => ay_data_in,
    O_DA                => ay_data_out,
    O_DA_OE_L           => ay_data_oe,
    -- control
    I_A9_L              => '0',
    I_A8                => '1',
    I_BDIR              => ay_bdir,
    I_BC2               => '1',
    I_BC1               => ay_bc1,
    I_SEL_L             => '1',

    O_AUDIO             => analog_sound,
    -- port a
    I_IOA               => ay_pa_in,	--joysticks
    O_IOA               => ay_pa_out,
    O_IOA_OE_L          => open,		--ioa_oe_l,
    -- port b
    I_IOB               => "00000000", 	--portb nc at AY38912
    O_IOB               => open, 		--portb nc at AY38912
    O_IOB_OE_L          => open, 		--portb nc at AY38912

    ENA                 => '1',
    RESET_L             => not cpu_rst,
    CLK                 => clk1m5
    );


x_int : signed_integ
	port map (
			clk_i      		=> cpu_clk,
			rst_i      		=> cpu_rst,
			ramp_i	   		=> not via_pb_out(7),
			zero_i	   		=> not via_ca2_out,
			data_i			=> analog_x,
			zeroref_i		=> analog_offs,
			data_o			=> analog_x_int,
			data_o_stb		=> cs_x_int
			);
			
y_int : signed_integ
	port map (
			clk_i      		=> cpu_clk,
			rst_i      		=> cpu_rst,
			ramp_i	   		=> not via_pb_out(7),
			zero_i	   		=> not via_ca2_out,
			data_i			=> analog_y,
			zeroref_i		=> analog_offs,
			data_o			=> analog_y_int,
			data_o_stb		=> cs_y_int
			);


----------------------------------------------------------------------
--
-- Process to decode memory map
--
----------------------------------------------------------------------

mem_decode: process(cpu_rst,cpu_data_out,resetsr,cpu_addr,cart_data_out,
					cpu_vma,sram_data_out,io_data_out,rom_data_out)
begin
--	if sys_clk'event and sys_clk='1' then

		  cpu_rst    <= resetsr(1); -- CPU reset is active high	  
		  ares		 <= cpu_areset;
		
		if cpu_rst = '1' then
		cpu_data_in <= (others=>'Z');
		  cpu_data_in <= cpu_data_out;
		  rom_cs      <= '0';
		  io_cs       <= '0';
		  sram_cs     <= '0';
		  cart_cs	  <= '0';      
		  cpu_halt   <= '0';
		  cpu_hold   <= cpu_ahold;
		  cpu_firq   <= '0';   
		  cpu_irq    <= '0';   
		  cpu_nmi    <= cpu_anmi;
--		  cpu_rst    <= resetsr(1); -- CPU reset is active high	  
		  else
		 

	-- Memory Decoder

		if cpu_addr( 15 ) = '0' then 				--CART 0x0000-0x7FFF
			cpu_data_in <= cart_data_out;
			cart_cs     <= cpu_vma;
		else
			cart_cs <= '0';
		end if;  

		if cpu_addr( 15 downto 11 ) = "11001" then 	--RAM 0xC800-0xCFFF
		--if cpu_addr( 15 downto 12 ) = "1100" then 	--RAM 0xC800-0xCFFF
			cpu_data_in <= sram_data_out;
			sram_cs     <= cpu_vma;
		else
			sram_cs <= '0';
		end if;
		
		if cpu_addr( 15 downto 12 ) = "1101" then 	--IO 0xD000-0xDFFF	    
			cpu_data_in <= io_data_out;
			io_cs     <= cpu_vma;
		else
			io_cs <= '0';	
		end if;

		if cpu_addr( 15 downto 13 ) = "111" then 	--ROM 0xE000-0xFFFF
			cpu_data_in <= rom_data_out;
			rom_cs      <= cpu_vma;
		else
			rom_cs <= '0';
		end if;
		
  end if;      
end process;


update_alg : process( clk6m, clk1m5 ,cpu_rst,via_pb_out,via_cb1_out,via_pb_oe,via_ca2_out,
					via_cb2_out,analog_vec,analog_y_temp,via_pa_oe,ay_data_oe,ay_data_out,
					analog_x,via_pa_out)

-- Resolution: 330  410

begin	
	if cpu_rst = '1' then
	
		ay_bc1		   			<= '0';
		ay_bdir        			<= '0';
		via_pa_in				<= "00000000";
		via_pb_in(6)			<= '1'; --cart input
	
	else
	
		
		via_pb_in(7 downto 6)	<= via_pb_out(7 downto 6);
		via_pb_in(5) <= compare_out;
		via_pb_in(4 downto 0)	<= via_pb_out(4 downto 0);
		via_cb1_in		<= via_cb1_out;
		via_cb1_in		<= '0';
		if via_pb_oe(3) = '1' then
			ay_bc1			<= via_pb_out(3);
		end if;
		if via_pb_oe(4) = '1' then
			ay_bdir        	<= via_pb_out(4);
		end if;
		via_ca2_in		<= via_ca2_out;
		cs_x 			<= not via_pb_out(7);
		cs_y 			<= not via_pb_out(7);
		via_cb2_in		<= via_cb2_out;
		
		--demux x channel
--		if (via_pb_out(0) = '1') and (via_pb_out(7) = '0') then
		if (via_pb_out(7) = '0') then
			analog_x 	<= analog_vec;
		else
			analog_x 	<= "00000000";
		end if;
	
		
		--demux y channel			
		if 	(via_pb_out(1) = '0') and (via_pb_out(2) = '0') and (via_pb_out(0) = '0') then
			 analog_y_temp 	<= analog_vec;
		end if;
		
		if (via_pb_out(7) = '0') then
			analog_y 	<= analog_y_temp;
		else
			analog_y 	<= "00000000";
		end if;
		
		--demux z channel
		if 	(via_pb_out(1) = '0') and (via_pb_out(2) = '1') and (via_pb_out(0) = '0') then
			cs_z <= '1';
			analog_z_sh <= analog_vec;
		else
			cs_z <= '0';
		end if;
		
		if via_cb2_out = '0' then
			analog_z <= "00000000";
		else
			analog_z <= analog_z_sh;
		end if;
		
		--demux offset channel
		if 	(via_pb_out(1) = '1') and (via_pb_out(2) = '0') and (via_pb_out(0) = '0') then
			analog_offs <= analog_vec;
		end if;
		
		-- mux comparator in
		if (via_pb_out(2 downto 1)="00") then
			comparator_in<=p1_x;
		elsif	(via_pb_out(2 downto 1)="01") then
			comparator_in<=p1_y;
		elsif	(via_pb_out(2 downto 1)="10") then
			comparator_in<=p2_x;
		else -- "11"
			comparator_in<=p2_y;
		end if;
		
		if via_pa_oe = "00000000" then					--Read/write PortA	
			if via_pb_out(3) = '1' then  
				if ay_data_oe = '0' then
					via_pa_in		<= ay_data_out;
					analog_vec 		<= ay_data_out;-- xor "10000000";
				else
					via_pa_in	    <= via_pa_out;
					analog_vec		 <= via_pa_out;-- xor "10000000";
				end if;
			else
				via_pa_in		<= analog_x;--via_pa_out;		--read PortA , DA hat kein Latch daher 1 (NMOS)
				analog_vec(7 downto 0) 		<= "01111111";
			end if;
		else					
			via_pa_in		<= via_pa_out;				--write PortA
			ay_data_in		<= via_pa_out;				
			analog_vec 		<= via_pa_out;-- xor "10000000";
		end if;		
	end if;
end process;


update_joys: process(cpu_clk) --Synchronous because we handle user input here.
begin
   if rising_edge(cpu_clk) and clk6m='1' then
		if comparator_in>analog_vec then 
			compare_out<='0';
		else
			compare_out<='1';
		end if;
	end if;
end process;

reset : process begin 
   wait until rising_edge(clk1m5);
   resetsr <= resetsr(0) & ares;
end process;

vex_clk : process (cpu_clk,clk6m,clk1m5)
begin
	if rising_edge(cpu_clk) then
		if cnt6m = 2-1 then
		   clk6m  <= '1';
		   cnt6m  <= 0;
		else
		   clk6m  <= '0';
		   cnt6m  <= cnt6m +1;
		end if;
	 end if; 
   
	if rising_edge(clk6m) then       
		if cnt1m5  = 4-1 then
		   clk1m5  <= '1';
		   cnt1m5  <= 0;
		else
		   clk1m5  <= '0';
		   cnt1m5  <= cnt1m5 +1;
		end if;
	 end if; 
	 
  if rising_edge(clk1m5) then       
		if cnt750k  = 2-1 then
		   clk750k  <= '1';
		   cnt750k  <= 0;
		else
		   clk750k  <= '0';
		   cnt750k  <= cnt750k +1;
		end if;
  end if;
  
end process;

analog_x_out<=analog_x_int;
analog_y_out<=analog_y_int;
analog_z_out<=analog_z;
analog_sound_out<=analog_sound;

via_wen<=not cpu_rw and io_cs and cpu_addr(12);
via_ren<=cpu_rw and io_cs and cpu_addr(12);

ay_pa_in<=p2_buttons&p1_buttons;

end Behaviour;

